%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) María Pérez Ortiz (i82perom at uco dot es)
% 
% This code implements the ordinal graph-based over-sampling via neighbourhood information using a probability function for the intra-class edges (OGO-NI method) presented in the paper Graph-Based Approaches for Over-sampling in the context of Ordinal Regression published in the IEEE Transactions on Knowledge and Data Engineering. 
% 
% The code has been tested with Ubuntu 12.04 x86_64 and Matlab R2009a
% 
% If you use this code, please cite the associated paper
% Code updates and citing information:
% http://www.uco.es/grupos/ayrna/GBOforOR
% 
% AYRNA Research group's website:
% http://www.uco.es/ayrna 
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 3
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
% Licence available at: http://www.gnu.org/licenses/gpl-3.0.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [newpatterns] = OGONI(dataset,index_class,k,n_new)
% Over-sample patterns using the OGO-NI method (ordinal graph-based over-sampling via neighbourhood information using a probability function for the intra-class edges).
% Input variables: 
% *dataset: train structure containing the train patterns and the associated
%  targets
% *index_class: Number that indicates the index of the class to be over-sampled
% *k: Number of neighbors to consider for constructing the graph
% *n_new: Number of new patterns to create
% Output variables
% *newpatterns: Output patterns from the over-sampling procedure

    previous_patterns = dataset.patterns(dataset.targets == index_class -1,:);
    class_for_oversampling = dataset.patterns(dataset.targets == index_class,:);
    next_patterns = dataset.patterns(dataset.targets == index_class +1,:);
    
    % Is it the first class?
    if sum(dataset.targets == (index_class - 1))==0,
        V1 = [];
        adjacency_G1 = [];
    else
        [D] = pdist2(previous_patterns,class_for_oversampling);
        D2 = D;
        adjacency_1 = zeros(size(class_for_oversampling,1),size(previous_patterns,1));
        adjacency_2 = zeros(size(class_for_oversampling,1),size(previous_patterns,1));

	% Compute the adjacency matrices
        for i=1:size(class_for_oversampling,1),
            for j=1:k,
                [M,idx] = min(D(:,i));
                D(idx,i) = inf;
                adjacency_1(i,idx) = 1;
            end
        end
        
	% Compute the number of neighbors
        neighbors = sum(adjacency_1,1);
	% Remove isolated patterns
        adjacency_1(:,neighbors==0) = [];
        adjacency_2(:,neighbors==0) = [];
        D2(neighbors==0,:) = [];
        previous_patterns = previous_patterns(neighbors>0,:);
        
        for i=1:size(previous_patterns,1),
            for j=1:k,
                [M,idx] = min(D2(i,:));
                D2(i,idx) = inf;
                adjacency_2(idx,i) = 1;
            end
        end

        adjacency_G1 = adjacency_1 .* adjacency_2;
        neighbors = sum(adjacency_G1,1);
        adjacency_G1(:,neighbors==0) = [];
        V1 = previous_patterns(neighbors>0,:);
    end
    
   
    if sum(dataset.targets == (index_class + 1))==0,
        V3 = [];
        adjacency_G3 = [];
    else
        [D] = pdist2(next_patterns,class_for_oversampling);
        D2 = D;
        adjacency_1 = zeros(size(class_for_oversampling,1),size(next_patterns,1));
        adjacency_2 = zeros(size(class_for_oversampling,1),size(next_patterns,1));

        for i=1:size(class_for_oversampling,1),
            for j=1:k,
                [M,idx] = min(D(:,i));
                D(idx,i) = inf;
                adjacency_1(i,idx) = 1;
            end
        end

        neighbors = sum(adjacency_1,1);
        adjacency_1(:,neighbors==0) = [];
        adjacency_2(:,neighbors==0) = [];
        D2(neighbors==0,:) = [];
        next_patterns = next_patterns(neighbors>0,:);
        
        for i=1:size(next_patterns,1),
            for j=1:k,
                [M,idx] = min(D2(i,:));
                D2(i,idx) = inf;
                adjacency_2(idx,i) = 1;
            end
        end
    adjacency_G3 = adjacency_1 .* adjacency_2;
    neighbors = sum(adjacency_G3,1);
    adjacency_G3(:,neighbors==0) = [];
    V3 = next_patterns(neighbors>0,:);
    end
    
    
    adjacency_G2 = zeros(size(class_for_oversampling,1),size(class_for_oversampling,1));
    
    % Compute distances
    [D] = pdist2(class_for_oversampling,class_for_oversampling);
    for i=1:size(class_for_oversampling,1),
        for j=1:k+1,
            [M,idx] = min(D(i,:));
            D(i,idx) = inf;
            adjacency_G2(i,idx) = 1;
        end
    end
    adjacency_G2 = adjacency_G2 - eye(size(adjacency_G2));
    % Compute the final graph
    adjacency_G = [adjacency_G2, adjacency_G1, adjacency_G3];
    patterns = [class_for_oversampling; V1; V3];
    targets = [ones(1,size(class_for_oversampling,1))*index_class, ones(1,size(V1,1))*(index_class-1), ones(1,size(V3,1))*(index_class+1)];
    
    % Find the transited patterns
    [rows,columns] = find(adjacency_G==1);
    
    % Gamma density function
    gamma_df = random('gam',2,0.15,[1,1500]);
    gamma_df = (gamma_df-min(gamma_df))./max(gamma_df);
    
    % Over-sample patterns
    for i=1:n_new,
        r = randi(numel(rows),1);
        if targets(rows(r))~=targets(columns(r)),
            aleatorio = randi(1500,1);
            if targets(rows(r))==index_class,
                newpatterns(i,:) = patterns(rows(r),:) + gamma_df(aleatorio)*(patterns(columns(r),:) - patterns(rows(r),:));
            else
                newpatterns(i,:) = patterns(columns(r),:) + gamma_df(aleatorio)*(patterns(rows(r),:) - patterns(columns(r),:));
            end
        else
            newpatterns(i,:) = patterns(rows(r),:) + rand()*(patterns(columns(r),:) - patterns(rows(r),:));
        end
    end
 
end
