%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Copyright (C) María Pérez Ortiz (i82perom at uco dot es)
% 
% This code implements the ordinal graph-based over-sampling via shortest paths using a probability function for the intra-class edges (OGO-SP method) presented in the paper Graph-Based Approaches for Over-sampling in the context of Ordinal Regression published in the IEEE Transactions on Knowledge and Data Engineering. 
% 
% The code has been tested with Ubuntu 12.04 x86_64 and Matlab R2009a
% 
% If you use this code, please cite the associated paper
% Code updates and citing information:
% http://www.uco.es/grupos/ayrna/GBOforOR
% 
% AYRNA Research group's website:
% http://www.uco.es/ayrna 
%
% This program is free software; you can redistribute it and/or
% modify it under the terms of the GNU General Public License
% as published by the Free Software Foundation; either version 3
% of the License, or (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
%
% You should have received a copy of the GNU General Public License
% along with this program; if not, write to the Free Software
% Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. 
% Licence available at: http://www.gnu.org/licenses/gpl-3.0.html
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [newpatterns] = OGOSP(dataset,class_index,k,n_new)
% Over-sample patterns using the OGO-SP method (ordinal graph-based over-sampling via shortest paths using a probability function for the intra-class edges).
% Input variables: 
% *dataset: train structure containing the train patterns and the associated
%  targets
% *index_class: Number that indicates the index of the class to be over-sampled
% *k: Number of neighbors to consider for constructing the graph
% *n_new: Number of new patterns to create
% Output variables
% *newpatterns: Output patterns from the over-sampling procedure

    % Select the patterns belonging to the previous and next class to the one we want to overs-sample 
    previous_patterns = dataset.patterns(dataset.targets == class_index -1,:);
    class_for_oversampling = dataset.patterns(dataset.targets == class_index,:);
    next_patterns = dataset.patterns(dataset.targets == class_index +1,:);
    
    extreme_class = 0;
   % Is it the first class?
    if sum(dataset.targets == (class_index - 1))==0,
        V1 = [];
        adjacency_G1 = [];
        D_1 = [];
	extreme_class = 1;
    else
        [D] = pdist2(previous_patterns,class_for_oversampling);
        D2 = D;
        D_1 = D;
        adjacency_1 = zeros(size(class_for_oversampling,1),size(previous_patterns,1));
        adjacency_2 = zeros(size(class_for_oversampling,1),size(previous_patterns,1));

	% Compute the adjacency matrices
        for i=1:size(class_for_oversampling,1),
            for j=1:k,
                [M,idx] = min(D(:,i));
                D(idx,i) = inf;
                adjacency_1(i,idx) = 1;
            end
        end

	% Compute the number of neighbors
        neighbors = sum(adjacency_1,1);
	% Remove isolated patterns
        adjacency_1(:,neighbors==0) = [];
        adjacency_2(:,neighbors==0) = [];
        D2(neighbors==0,:) = [];
        D_1(neighbors==0,:)  = [];
        previous_patterns = previous_patterns(neighbors>0,:);
        
        for i=1:size(previous_patterns,1),
            for j=1:k,
                [M,idx] = min(D2(i,:));
                D2(i,idx) = inf;
                adjacency_2(idx,i) = 1;
            end
        end

        adjacency_G1 = adjacency_1 .* adjacency_2;
        neighbors = sum(adjacency_G1,1);
        adjacency_G1(:,neighbors==0) = [];
        D_1(neighbors==0,:) = [];
        V1 = previous_patterns(neighbors>0,:);
    end
    
   
    if sum(dataset.targets == (class_index + 1))==0,
        V3 = [];
        adjacency_G3 = [];
        D_3 = [];
	extreme_class = 1;
    else
        [D] = pdist2(next_patterns,class_for_oversampling);
        D2 = D;
        D_3 = D;
        adjacency_1 = zeros(size(class_for_oversampling,1),size(next_patterns,1));
        adjacency_2 = zeros(size(class_for_oversampling,1),size(next_patterns,1));

        for i=1:size(class_for_oversampling,1),
            for j=1:k,
                [M,idx] = min(D(:,i));
                D(idx,i) = inf;
                adjacency_1(i,idx) = 1;
            end
        end

        neighbors = sum(adjacency_1,1);
        adjacency_1(:,neighbors==0) = [];
        adjacency_2(:,neighbors==0) = [];
        D2(neighbors==0,:) = [];
        D_3(neighbors==0,:) = [];
        next_patterns = next_patterns(neighbors>0,:);
        
        for i=1:size(next_patterns,1),
            for j=1:k,
                [M,idx] = min(D2(i,:));
                D2(i,idx) = inf;
                adjacency_2(idx,i) = 1;
            end
        end
    adjacency_G3 = adjacency_1 .* adjacency_2;
    neighbors = sum(adjacency_G3,1);
    adjacency_G3(:,neighbors==0) = [];
    D_3(neighbors==0,:) = [];
    V3 = next_patterns(neighbors>0,:);
    end
    
    
    adjacency_G2 = zeros(size(class_for_oversampling,1),size(class_for_oversampling,1));
    
    % Compute distances
    [D] = pdist2(class_for_oversampling,class_for_oversampling);
    D_2 = D;
    for i=1:size(class_for_oversampling,1),
        for j=1:k+1,
            [M,idx] = min(D(i,:));
            D(i,idx) = inf;
            adjacency_G2(i,idx) = 1;
        end
    end
    adjacency_G2 = adjacency_G2 - eye(size(adjacency_G2));

    % Compute the final graph
    adjacency_G = [adjacency_G2, adjacency_G1, adjacency_G3];
    patterns = [class_for_oversampling; V1; V3];
    distances = [D_2; D_1; D_3];
    distances(distances==0) = NaN;
    distances = distances .* adjacency_G';

    % Set the distance matrix
    set_of_distances = zeros(size(patterns,1),size(patterns,1));
    set_of_distances(1:end,1:size(class_for_oversampling,1)) = distances;
    set_of_distances(1:size(class_for_oversampling,1),1:end) = distances';
    targets = [ones(1,size(class_for_oversampling,1))*class_index, ones(1,size(V1,1))*(class_index-1), ones(1,size(V3,1))*(class_index+1)];
    
    % We need to know how many times each pattern is transited when walking through the different paths
    transited = zeros(size(patterns,1),size(patterns,1));
    if sum(dataset.targets == (class_index - 1))==0,
       for i=1:size(class_for_oversampling,1),
            for j=1:size(V3,1),
                [d_graph, c_graph] = dijkstra(set_of_distances,i,size(class_for_oversampling,1)+j);
                for z=1:numel(c_graph)-1,
                        transited(c_graph(z),c_graph(z+1)) = transited(c_graph(z),c_graph(z+1)) + 1;
                        transited(c_graph(z+1),c_graph(z)) = transited(c_graph(z+1),c_graph(z)) + 1;
                end
            end
        end 
    elseif sum(dataset.targets == (class_index + 1))==0,
        for i=1:size(class_for_oversampling,1),
            for j=1:size(V1,1),
                [d_graph, c_graph] = dijkstra(set_of_distances,i,size(class_for_oversampling,1)+j);
                for z=1:numel(c_graph)-1,
                        transited(c_graph(z),c_graph(z+1)) = transited(c_graph(z),c_graph(z+1)) + 1;
                        transited(c_graph(z+1),c_graph(z)) = transited(c_graph(z+1),c_graph(z)) + 1;
                end
            end
        end 
    else
        for i=1:size(V1,1),
            for j=1:size(V3,1),
                [d_graph, c_graph] = dijkstra(set_of_distances,size(class_for_oversampling,1)+i,size(class_for_oversampling,1)+size(V1,1)+j);
                for z=1:numel(c_graph)-1,
                        transited(c_graph(z),c_graph(z+1)) = transited(c_graph(z),c_graph(z+1)) + 1;
                        transited(c_graph(z+1),c_graph(z)) = transited(c_graph(z+1),c_graph(z)) + 1;
                end
            end
        end
    end
    
    % Find the transited patterns
    [rows,columns] = find(triu(transited)~=0);

    % Gamma density function
    gamma_df = random('gam',2,0.15,[1,1500]);
    gamma_df = (gamma_df-min(gamma_df))./max(gamma_df);
    
% Over-sample patterns
if extreme_class == 0,
     for i=1:n_new,
         r = randi(numel(rows),1);
         if targets(rows(r))~=targets(columns(r)),
            random_number = randi(1500,1);
            if targets(rows(r))==class_index,
                newpatterns(i,:) = patterns(rows(r),:) + gamma_df(random_number)*(patterns(columns(r),:) - patterns(rows(r),:));
            else
                newpatterns(i,:) = patterns(columns(r),:) + gamma_df(random_number)*(patterns(rows(r),:) - patterns(columns(r),:));
            end
         else
            newpatterns(i,:) = patterns(rows(r),:) + rand()*(patterns(columns(r),:) - patterns(rows(r),:));
         end
     end
else
        for i=1:n_new,
            r = randi(numel(rows),1);
            if targets(rows(r))==targets(columns(r)),
                newpatterns(i,:) = patterns(rows(r),:) + rand()*(patterns(columns(r),:) - patterns(rows(r),:));
            else
		random_number = randi(1500,1);
		if rand()>0.5
		        if targets(rows(r))==class_index,
		        	newpatterns(i,:) = patterns(rows(r),:) + gamma_df(random_number)*(patterns(columns(r),:) - patterns(rows(r),:));
		        else
		        	newpatterns(i,:) = patterns(columns(r),:) + gamma_df(random_number)*(patterns(rows(r),:) - patterns(columns(r),:));
		    	end
		else
                	if targets(rows(r))==class_index,
		        	newpatterns(i,:) = patterns(rows(r),:) - gamma_df(random_number)*(patterns(columns(r),:) - patterns(rows(r),:));
		        else
		        	newpatterns(i,:) = patterns(columns(r),:) - gamma_df(random_number)*(patterns(rows(r),:) - patterns(columns(r),:));
		    	end
		end
            end
        end
    end



end
